package server

import (
	"errors"
	"sync"
	"testing"
	"time"

	"github.com/centrifugal/centrifugo/libcentrifugo/conns"
	"github.com/gorilla/websocket"
	"github.com/stretchr/testify/assert"
)

type testWSConnection struct {
	mu         sync.RWMutex
	writeErr   bool
	readErr    bool
	controlErr bool
	closeErr   bool
	closed     bool
}

func (c *testWSConnection) ReadMessage() (messageType int, p []byte, err error) {
	if c.readErr {
		return websocket.TextMessage, nil, errors.New("error")
	}
	return websocket.TextMessage, []byte("test"), nil
}

func (c *testWSConnection) WriteMessage(messageType int, data []byte) error {
	if c.writeErr {
		return errors.New("error")
	}
	return nil
}

func (c *testWSConnection) WritePreparedMessage(*websocket.PreparedMessage) error {
	if c.writeErr {
		return errors.New("error")
	}
	return nil
}

func (c *testWSConnection) WriteControl(messageType int, data []byte, deadline time.Time) error {
	if c.controlErr {
		return errors.New("error")
	}
	return nil
}

func (c *testWSConnection) SetWriteDeadline(t time.Time) error {
	return nil
}

func (c *testWSConnection) EnableWriteCompression(enabled bool) {
	return
}

func (c *testWSConnection) Close() error {
	c.mu.Lock()
	defer c.mu.Unlock()
	c.closed = true
	if c.closeErr {
		return errors.New("error")
	}
	return nil
}

func TestWSConnPing(t *testing.T) {
	ws := &testWSConnection{}
	c := newWSSession(ws, 1*time.Nanosecond, 0, 0)
	c.ping()
	conn := c.ws.(*testWSConnection)
	conn.mu.Lock()
	assert.Equal(t, false, conn.closed)
	conn.mu.Unlock()
}

func TestWSConnPingFailed(t *testing.T) {
	ws := &testWSConnection{controlErr: true}
	c := newWSSession(ws, 1*time.Nanosecond, 0, 0)
	c.ping()
	conn := c.ws.(*testWSConnection)
	conn.mu.Lock()
	assert.Equal(t, true, conn.closed)
	conn.mu.Unlock()
}

func TestWSConnPingAfterClose(t *testing.T) {
	ws := &testWSConnection{}
	c := newWSSession(ws, 1*time.Nanosecond, 0, 0)
	err := c.Close(conns.DefaultDisconnectAdvice)
	assert.Equal(t, nil, err)
	c.ping()
	conn := c.ws.(*testWSConnection)
	conn.mu.Lock()
	assert.Equal(t, true, conn.closed)
	conn.mu.Unlock()
}

func TestSendAfterClose(t *testing.T) {
	ws := &testWSConnection{}
	c := newWSSession(ws, 1*time.Nanosecond, 0, 0)
	err := c.Close(conns.DefaultDisconnectAdvice)
	assert.Equal(t, nil, err)
	err = c.Send(conns.NewQueuedMessage([]byte("test"), false))
	assert.Equal(t, nil, err)
}
